// TFormEx 0.9 (beta) by Fred Ackers (aqiruse@softhome.net)

unit FormEx;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, contnrs, Extctrls, jpeg;

type
  // this is for desktop docking
  TDockType = (dtLeft, dtRight, dtTop, dtBottom, dtNone);
  TCaptionBars = (dtNoCaptionBar, dtNormal);
  TMoveType = (dtClientClick, dtMoveCaptionBarOnly);
  TBackType = (baNormal, baGradient, baBackImage);
  TImageMethod = (baTile, baStretch, daDraw);
  TImageType = (JPG, BMP);
  // For Form to Form docking
  TProps = set of (dt_LeftSide, dt_RightSide, dt_TopSide,
        dt_BottomSide, dt_DockToTop, dt_DockToLeft, dt_DockToRight, dt_DockToBottom,
        dt_Single,dt_Permanent);
  (* Docking parameters
    // Sides Match
  These will get both left sides at the same point, both right at the same point, etc.
  dt_LeftSide = 2; // Both at 0
  dt_TopSide = 4; // both at 0
  dt_RightSide = 8; // both at 800
  dt_BottomSide = 16; // both at 600

  // opposites are docked
  dt_DockToTop = 24; // Bottom is docked to top of other form
  dt_DockToLeft = 48; // Left is docked to right of other form
  dt_DockToRight = 96; // right is docked to Left of other form
  dt_DockToBottom = 192; // top is docked to bottom of other form

  // Single Click docks or undocks
  dt_single
  // Automatically permanent
  dt_Permanent = 1536; // stay permanently docked *)

  TMoveWndType = (mvBoth, mvNone);
const
  MAXWNDS = 20;
type
  TFormEx = class(TComponent)
  private
   FFormProp : array [0..MAXWNDS] of TProps;
   FFormMove : TMoveType;
   FMinWidth : LongInt;
   FMinHeight : LongInt;
   FMaxWidth : LongInt;
   FMaxHeight : LongInt;
   FHasMinWidth : Boolean;
   FHasMinHeight : Boolean;
   FHasMaxWidth : Boolean;
   FHasMaxHeight : Boolean;
   FHideCaptionBar : Boolean;
   FSnapToDesktopSides : Boolean;
   FDockSidesOnly : Boolean;
   FDockType : TDockType;
   FDockWidth : Boolean;
   FDockHeight : Boolean;
   FTransparent : Boolean;
   FCaptionBar : TCaptionBars;
   FBackground : string;
   FFromColor : TColor;
   FToColor : TColor;
   FBackType : TBackType;
   FFormList : TComponentList;
   FImageMethod : TImageMethod;
   FImageType : TImageType;
   FIsDocked : array [0 ..MAXWNDS] of boolean;
   FMoveBoth : TMoveWndType;
   OverrideWndProc : Pointer;
   FormWndProc : Pointer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;

    procedure WMNCHitTest(var M: TWMNCHitTest); message wm_NCHitTest;
    procedure WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo); message WM_GETMINMAXINFO;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMMove(var Message: TWMMove); message WM_MOVE;

    procedure WndProcOverride(var Msg: TMessage);

    procedure MakeTransparent;
    procedure SnapToSides;
    procedure DockForm(dtType : TDockType; DockWidth : boolean = false; DockHeight : boolean = false);
    procedure SetCaptionBar;
    procedure DoDocking;
    Function GetForm(index : integer) : TComponent;
    procedure AddForm(index : integer; Value : TComponent; Properties : TProps);
    Function IsClose(index : integer) : boolean;
    procedure SetTransparent(value : boolean);
    procedure Refresh;
   protected
    function GetNewColor(Color1,Color2:TColor;Num:Integer):TColor;
   published
    // Bitmap to place on back of form
    property Background : string read FBackground write FBackground;
    // Image type
    property ImageType : TImageType read FImageType write FImageType;
    // Drawing method
    property ImageMethod : TImageMethod read FImageMethod write FImageMethod;

    // 1st Gradient color for background
    property GradientFromColor : TColor read FFromColor write FFromColor default clWhite;
    // Color to draw gradient to
    property GradientToColor : TColor read FToColor write FToColor default clBlue;

    // Toggle between background options
    property BackgroundType : TBackType read FBackType write FBackType;

    // dtNoCaptionBars Removes caption bar if desired, dtNormal shows them as you set them
    property CaptionBar : TCaptionBars read FCaptionBar write FCaptionBar;

    //Set window to become transparent except for controls
    property Transparent : boolean read FTransparent write SetTransparent default false;

    // dtClientClick allows you to move by clicking anywhere on the for,
    // otherwise moving is done by clicking on caption bar only as usual
    property AllowMoveByClick: TMoveType read FFormMove write FFormMove;

    // Move both windows at once when they are docked
    property MoveBothWnds : TMoveWndType read FMoveBoth write FMoveBoth;
    // Sets minimum an maximum sizes for the form
    property MinWidth : Longint read FMinWidth write FMinWidth default 100;
    property MinHeight : Longint read FMinHeight write FMinHeight default 100;
    property MaxWidth : Longint read FMaxWidth write FMaxWidth default 1000;
    property MaxHeight : Longint read FMaxHeight write FMaxHeight default 1000;
    // Set True for which values you want to use
    property HasMinWidth : Boolean read FHasMinWidth write FHasMinWidth default false;
    property HasMinHeight : Boolean read FHasMinHeight write FHasMinHeight default false;
    property HasMaxWidth : Boolean read FHasMaxWidth write FHasMaxWidth default false;
    property HasMaxHeight : Boolean read FHasMaxHeight write FHasMaxHeight default false;

    // Is true, when moving a form and the form is moved off of the desktop,
    // then this will reset the form so that it does not move off the desktop
    property SnapToDesktopSides : Boolean read FSnapToDesktopSides write FSnapToDesktopSides default false;

    // If true this will dock the form to the sides of the desktop much like the taskbar
    property DockToSidesOnly : Boolean read FDockSidesOnly write FDockSidesOnly default false;
    // this tells the form which side to dock to if any
    property DockType : TDockType read FDockType write FDockType default dtNone;
    // if true then the form resizes to the desktop width
    property DockingTakesWidth : Boolean read FDockWidth write FDockWidth;
    // if true then form resizes to desktop height
    property DockingTakesHeight : Boolean read FDockHeight write FDockHeight;
  end;

procedure Register;

implementation

procedure TFormEx.SetTransparent(Value : boolean);
begin
  FTransparent := value;
  Refresh;
end;

Function TFormEx.IsClose(index : integer) : boolean;
var
  closest : array[0..1] of longint;
begin
  Result := false;
  if (dt_LeftSide in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Left;
    closest[1] := TForm(Owner).Left;
    if (Low(closest) < 10) then Result := true;
  end;
  if (dt_TopSide in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Top;
    closest[1] := TForm(Owner).Top;
    if (Low(closest) < 10) then Result := true;
  end;
  if (dt_RightSide in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Left;
    closest[1] := (TForm(Owner).Left+
            TForm(Owner).Width)-(FFormList.Items[index] as TWinControl).Width;
    if (Low(closest) < 10) then Result := true;
  end;
  if (dt_BottomSide in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Top;
    closest[1] := (TForm(Owner).Top+
            TForm(Owner).Height)-(FFormList.Items[index] as TWinControl).Height;
    if (Low(closest) < 10) then Result := true;
  end;

  if (dt_DockToLeft in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Left;
    closest[1] := TForm(Owner).Left+TForm(Owner).Width;
    if (Low(closest) < 10) then Result := true;
  end
  else if (dt_DockToTop in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Top;
    closest[1] := TForm(Owner).Top+
            (FFormList.Items[index] as TWinControl).Height;
    if (Low(closest) < 10) then Result := true;
  end
  else if (dt_DockToRight in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Left;
    closest[1] := (TForm(Owner).Left+
            TForm(Owner).Width)-(FFormList.Items[index] as TWinControl).Width;
    if (Low(closest) < 10) then Result := true;
  end
  else if (dt_DockToBottom in FFormProp[index]) then
  begin
    closest[0] := (FFormList.Items[index] as TWinControl).Top;
    closest[1] := (TForm(Owner).Top+
            TForm(Owner).Height)-(FFormList.Items[index] as TWinControl).Height;
    if (Low(closest) < 10) then Result := true;
  end;
end;

Function TFormEx.GetForm(index : integer) : TComponent;
begin
  Result := FFormList.items[index];
end;

procedure TFormEx.AddForm(index : integer; Value : TComponent; Properties : TProps);
begin
  FFormList.Insert(index, Value);
  FFormProp[index] := Properties;
end;

procedure TFormEx.SetCaptionBar;
begin
  if (FHideCaptionBar = true) or (FCaptionBar = dtNoCaptionBar) then
  SetWindowLong(TForm(Owner).Handle, GWL_STYLE,
                 (GetWindowLong(TForm(Owner).Handle,GWL_STYLE)
                 or WS_POPUP) and not WS_DLGFRAME and not WS_CAPTION);
end;

procedure TFormEx.WMMove(var Message: TWMMove);
var
  Screen : TScreen;
begin
  TForm(Owner).Width := TForm(Owner).Width + 1;
  SnapToSides;
  if (DockToSidesOnly = true) then
    DockForm(DockType);
  if (Transparent = true) then
    MakeTransparent;
  DoDocking;
end;

procedure TFormEx.DoDocking;
var
  i : integer;
begin
  if (FFormList.Count > 0) and (FMoveBoth = mvBoth) then
  begin
    for i := 0 to FFormList.Count-1 do
    begin
      if (FIsDocked[i] = false) then continue;
      if (dt_LeftSide in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Left := TForm(Owner).Left;
      if (dt_TopSide in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Top := TForm(Owner).Top;
      if (dt_RightSide in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Left := (TForm(Owner).Left+
                TForm(Owner).Width)-(FFormList.Items[i] as TWinControl).Width;
      if (dt_BottomSide in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Top := (TForm(Owner).Top+
                TForm(Owner).Height)-(FFormList.Items[i] as TWinControl).Height;

      if (dt_DockToLeft in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Left := TForm(Owner).Left+TForm(Owner).Width
      else if (dt_DockToTop in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Top := TForm(Owner).Top+TForm(Owner).Height
      else if (dt_DockToRight in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Left := TForm(Owner).Left-
                (FFormList.Items[i] as TWinControl).Width
      else if (dt_DockToBottom in FFormProp[i]) then
        (FFormList.Items[i] as TWinControl).Top := TForm(Owner).Top-
                (FFormList.Items[i] as TWinControl).Height;
    end;
  end
  else if (FFormList.Count > 0) and (FMoveBoth = mvNone) then
  begin
    for i := 0 to FFormList.Count-1 do
    begin
      if (FIsDocked[i] = false) then continue;
      if (dt_LeftSide in FFormProp[i]) then
        TForm(Owner).Left := (FFormList.Items[i] as TWinControl).Left;
      if (dt_TopSide in FFormProp[i]) then
        TForm(Owner).Top := (FFormList.Items[i] as TWinControl).Top;
      if (dt_RightSide in FFormProp[i]) then
        TForm(Owner).Left := ((FFormList.Items[i] as TWinControl).Left+
                (FFormList.Items[i] as TWinControl).Width)-TForm(Owner).Width;
      if (dt_BottomSide in FFormProp[i]) then
        TForm(Owner).Top := ((FFormList.Items[i] as TWinControl).Top+
                (FFormList.Items[i] as TWinControl).Height)-TForm(Owner).Height;

      if (dt_DockToLeft in FFormProp[i]) then
        TForm(Owner).Left := (FFormList.Items[i] as TWinControl).Left-TForm(Owner).Width
      else if (dt_DockToTop in FFormProp[i]) then
        TForm(Owner).Top := (FFormList.Items[i] as TWinControl).Top-TForm(Owner).Height
      else if (dt_DockToRight in FFormProp[i]) then
        TForm(Owner).Left := (FFormList.Items[i] as TWinControl).Left+
                (FFormList.Items[i] as TWinControl).Width-TForm(Owner).Width
      else if (dt_DockToBottom in FFormProp[i]) then
        TForm(Owner).Top := (FFormList.Items[i] as TWinControl).Top+(FFormList.Items[i] as TWinControl).Height;
    end;
  end;
end;

procedure TFormEx.WMNCHitTest(var M: TWMNCHitTest);
var
  i : integer;
  reset : boolean;
begin
  inherited;
  if (FFormMove <> dtClientClick) then exit;
  if  M.Result = htClient then
    M.Result := htCaption;

  for i := 0  to FFormList.Count-1 do
  begin
    reset := false;
    if not (dt_Permanent in FFormProp[i]) and (FIsDocked[i] = true) and (dt_Single in FFormProp[i])then
    begin
      FIsDocked[i] := false;
      reset := true;
    end;
    if not (dt_Permanent in FFormProp[i]) and (FIsDocked[i] = false) and (dt_Single in FFormProp[i]) and
     (reset = false) then
      if (IsClose(i) = true) then
        FIsDocked[i] := true;
  end;
end;

procedure TFormEx.WmGetMinMaxInfo(var Msg: TWMGetMinMaxInfo);
var
  minmax: PMinMaxInfo;
begin
  minmax := Msg.MinMaxInfo;
  SetCaptionBar;
  if (HasMinWidth = true) then
    minmax.ptMinTrackSize.x := FMinWidth;
  if (HasMinHeight = true) then
    minmax.ptMinTrackSize.y := FMinHeight;
  if (HasMaxWidth = true) then
    minmax.ptMaxTrackSize.x := FMaxWidth;
  if (HasMaxHeight = true) then
    minmax.ptMaxTrackSize.y := FMaxHeight;

  Msg.Result := 0;
end;

destructor TFormEx.Destroy;
begin
  if Assigned(OverrideWndProc) then
    FreeObjectInstance(OverrideWndProc);

  inherited;
end;

constructor TFormEx.Create(AOwner: TComponent);
var
Region1 : array of tPoint;
Region1hrgn : hRgn;
i : integer;
Begin
  inherited;
  FFormList := TComponentList.Create;
  for i := 0 to MAXWNDS do
    FIsDocked[i] := true;
  if not(csDesigning in ComponentState) then
  begin
    OverrideWndProc := MakeObjectInstance(WndProcOverride);
    FormWndProc := Pointer(SetWindowLong(TForm(Owner).Handle, gwl_WndProc, Longint(OverrideWndProc)));
  end
  else
  begin
    OverrideWndProc := nil;
    FormWndProc := nil;
  end;
end;

procedure TFormEx.MakeTransparent;
begin
  if (Transparent = true) then
  begin
    TForm(Owner).Brush.Style:=bsClear;
    TForm(Owner).BorderStyle:=bsNone;
  end
  else
  begin
    TForm(Owner).Brush.Style:=bsSolid;
  end;
end;

function TFormEx.GetNewColor(Color1,Color2:TColor;Num:Integer):TColor;
var
  R,G,B          : Byte;
begin
      R :=  GetRValue(ColorToRGB(Color1)) + MulDiv (Num,
              GetRValue(ColorToRGB(Color2))-GetRValue(ColorToRGB(Color1)), TForm(Owner).Width - 1);
      G :=  GetGValue(ColorToRGB(Color1)) + MulDiv (Num,
              GetGValue(ColorToRGB(Color2))-GetGValue(ColorToRGB(Color1)), TForm(Owner).Width - 1);
      B :=  GetBValue(ColorToRGB(Color1)) + MulDiv (Num,
              GetBValue(ColorToRGB(Color2))-GetBValue(ColorToRGB(Color1)), TForm(Owner).Width - 1);
  Result:=RGB(R, G, B);
end;

procedure TFormEx.WMPaint(var Message: TWMPaint);
Var
  i : integer;
  img : TBitmap;
  jpeg : TJPegImage;
  ARect : TRect;
begin
  SnapToSides;
  if (DockToSidesOnly = true) then
    DockForm(DockType);
  MakeTransparent;

  inherited;
  if (FBackType = baGradient) then
  begin
    for i:=1 to TForm(Owner).Width do
    begin
      TForm(Owner).Canvas.Pen.Color:=GetNewColor(ColorToRGB(FFromColor),ColorToRGB(FToColor),i);
      TForm(Owner).Canvas.MoveTo(i,1);
      TForm(Owner).Canvas.LineTo(i,TForm(Owner).Height);
    end;
  end;
  if (FBackType = baBackImage) and (FBackground <> '') then
  begin
    ARect.Left := 0;
    ARect.Top := 0;
    ARect.Right := TForm(Owner).Width;
    ARect.Bottom := TForm(Owner).Height;
    if (ExtractFileExt(FBackground) = '.bmp') then
      FImageType := BMP;
    if (ExtractFileExt(FBackground) = '.jpg') then
      FImageType := JPG;
    if (FImageMethod = baTile) then //tile bitmap
      begin
        if (FImageType = JPG) then
        begin
          jpeg := TJpegImage.Create;
          jpeg.LoadFromFile(FBackground);
          Img := TBitmap.Create;
          img.Assign(jpeg);
          TForm(Owner).Canvas.Brush.Bitmap := Img;
          TForm(Owner).Canvas.FillRect(ARect);
          Img.Free;
          jpeg.Free;
        end
        else
        begin
          Img := TBitmap.Create;
          Img.LoadFromFile(FBackground);
          TForm(Owner).Canvas.Brush.Bitmap := Img;
          TForm(Owner).Canvas.FillRect(ARect);
          Img.Free;
        end;
      end else if (FImageMethod = baStretch) then//Stretch across the canvas.
      begin
        if (FImageType = JPG) then
        begin
          jpeg := TJpegImage.Create;
          jpeg.LoadFromFile(FBackground);
          Img := TBitmap.Create;
          img.Assign(jpeg);
          TForm(Owner).Canvas.Brush.Bitmap := Img;
          TForm(Owner).Canvas.StretchDraw(ARect, TForm(Owner).Canvas.Brush.Bitmap);
          Img.Free;
          jpeg.Free;
        end
        else
        begin
          Img := TBitmap.Create;
          Img.LoadFromFile(FBackground);
          TForm(Owner).Canvas.Brush.Bitmap := Img;
          TForm(Owner).Canvas.StretchDraw(ARect, TForm(Owner).Canvas.Brush.Bitmap);
          Img.Free;
        end;
      end
      else
      begin
        if (FImageType = JPG) then
        begin
          jpeg := TJpegimage.Create;
          jpeg.LoadFromFile(FBackground);
          Img := TBitmap.Create;
          Img.Assign(jpeg);
          TForm(Owner).Canvas.Brush.Bitmap := Img;
          TForm(Owner).Canvas.Draw(TForm(Owner).Left, TForm(Owner).Top, TForm(Owner).Canvas.Brush.Bitmap);
          Img.Free;
        end
        else
        begin
          Img := TBitmap.Create;
          Img.LoadFromFile(FBackground);
          TForm(Owner).Canvas.Brush.Bitmap := Img;
          TForm(Owner).Canvas.Draw(TForm(Owner).Left, TForm(Owner).Top, TForm(Owner).Canvas.Brush.Bitmap);
          Img.Free;
        end;
      end;
  end;
  DoDocking;
end;

procedure TFormEx.SnapToSides;
var
  Screen : TScreen;
begin
  if (FSnapToDesktopSides = true) then
  begin
    if (TForm(Owner).Top < Screen.DesktopTop) then
      TForm(Owner).Top := Screen.DesktopTop;

    if (TForm(Owner).Left < Screen.DesktopLeft) then
      TForm(Owner).Left := Screen.DesktopLeft;

    if (TForm(Owner).Top+TForm(Owner).Height > Screen.DesktopTop+Screen.DesktopHeight) then
      TForm(Owner).Height := Screen.DesktopHeight-TForm(Owner).Top;

    if (TForm(Owner).Left+TForm(Owner).Width > Screen.DesktopLeft+Screen.DesktopWidth) then
      TForm(Owner).Width := Screen.DesktopWidth-TForm(Owner).Left;

    if (TForm(Owner).Width > Screen.DesktopWidth) then
      TForm(Owner).Width := Screen.DesktopWidth;
    if (TForm(Owner).Height > Screen.DesktopHeight) then
      TForm(Owner).Height := Screen.DesktopHeight;
  end;
end;

procedure TFormEx.DockForm(dtType : TDockType; DockWidth : boolean; DockHeight : boolean);
var
  Screen : TScreen;
begin
  if (DockToSidesOnly = true) then
  begin
    if (dtType = dtTop) then
      TForm(Owner).Top := Screen.DesktopTop;

    if (dtType = dtLeft) then
      TForm(Owner).Left := Screen.DesktopLeft;

    if (dtType = dtBottom) then
      TForm(Owner).Top := Screen.DesktopHeight-TForm(Owner).Height;

    if (dtType = dtRight) then
      TForm(Owner).Left := Screen.DesktopWidth-TForm(Owner).Width;

    if (FDockWidth = true) then
    begin
      TForm(Owner).Width := Screen.DesktopWidth;
      TForm(Owner).Left := Screen.DesktopLeft;
    end;
    if (FDockHeight = true) then
    begin
      TForm(Owner).Height := Screen.DesktopHeight;
      TForm(Owner).Top := Screen.DesktopTop;
    end;
  end;
end;

procedure TFormEx.WndProcOverride(var Msg: TMessage);
begin
  if (Msg.Msg = WM_GETMINMAXINFO) then
    WMGetMinMaxInfo(TWMGetMinMaxInfo(Msg));
  if (Msg.Msg = WM_PAINT) then
    WMPaint(TWMPaint(Msg));
  if (Msg.Msg = WM_MOVE) then
    WMMove(TWMMove(Msg));

  Msg.Result := CallWindowProc(FormWndProc, TForm(Owner).Handle, Msg.Msg, Msg.WParam, Msg.LParam);
  if (Msg.Msg = wm_NCHitTest) then
    WMNCHitTest(TWMNCHitTest(Msg));
end;

procedure TFormEx.Refresh;
var
  Msg : TMessage;
begin
  Msg.Msg := WM_PAINT;
  Msg.WParam := 0;
  Msg.LParam := 0;
  WndProcOverride(Msg);
end;

procedure Register;
begin
  RegisterComponents('Standard', [TFormEx]);
end;

end.
